/*
Program: elim
File: elim.cpp
Author: Martin Voros
*/

#include <iostream>
#include <fstream>
#include <vector>
#include <iterator>
#include <getopt.h>

#include "elim.h"
#include "hash.h"

/*Global vars*/
char * output=NULL;
char * input=NULL;
bool verbose;
unsigned long ht_size=HT_DEF_SIZE;
unsigned long nin=0, nout=0;
int maxin=0, maxout=0;
unsigned long necc_eqs=0;

// Set & vector of all variables
set<string> svars;
vector<string> vars;
vector<bool> occured;
// Equations
vector<mpol_t> eqs;
matrix_t M;
vector<coeff_t> R;
// List of all monomials
vector<mpol_t::iterator> lmono;

using namespace std;

int 
main (int argc, char**argv) 
{
	/* Parsing input parameters */
	arguments(argc, argv);
	load_eq(input, eqs);

	/* Gaussian elimination */
	linearize(eqs);
	if (verbose) {
		cout << M;
		cout << "Vector R:" << endl;
		for (int i=0; i<R.size(); i++)
			cout << R.at(i) << " ";
 		cout << endl;
	}
	g2elim(M,R);
	if (verbose) {
		cout << "Linearized system" << endl;
		cout << M;
		cout << "Vector R:" << endl;
		for (int i=0; i<R.size(); i++)
			cout << R.at(i) << " ";
 		cout << endl;
	}
	

	/* Output */
	print_res(output);

	cout << "First " << necc_eqs << " eqs contain all "<< vars.size() << " variables." << endl;
	cout << "# of + on input: " << nin << endl;
	cout << "# of + on output: " << nout << endl;
	cout << "q = " << (float) nin/nout << endl;
	cout << "max # of + on input: " << maxin << endl;
	cout << "max # of + on input: " << maxout << endl;
	cout << "q_max = " << (float) maxin/maxout << endl;

	return 0;
}

/*
 * Returns true iff all elements of vector are true
 */
bool
really_true(vector<bool> &v)
{
	for (int i=0; i<v.size(); i++) 
		if (!v.at(i))
			return false;
	
	return true;
}

/*
 * Returns index of monomial in list of the monos else -1
 */
long
mono_member(mpol_t::iterator mono)
{
	for (int i=0; i<lmono.size(); i++) {
		if (*mono == *lmono.at(i)) {
			return i;
		}
	}
	return -1;
}

/*
 * Returns degree of monomial
 */
int
deg(mpol_t::iterator mono)
{
	long deg=0;

	for (long i=0; i<mono->size(); i++) 
		deg+=(*mono)[i];

	return deg;
}


/*
 * Get list of all monomials
 */
HASH_CELL **
get_monomial(vector<mpol_t> &eqs)
{
	HASH_CELL ** table;
	unsigned long key;
	unsigned long id=1;
	
	table=ht_alloc(ht_size);
	for (int i=0; i<eqs.size(); i++) {
		for(mpol_t::iterator it=eqs.at(i).begin(); 
			it !=eqs.at(i).end(); it++) {
			if (deg(it)>1) {
				key=hash_mono(it,ht_size);
				if ( 0 == hash_lookup(it, table[key]) ) {
					table[key]=insert_mono(it, table[key], lmono, id);
					id++;
				}
			}
		}	
	}

	return table;
}

/*
 * Linearize system of equations
 */
void
linearize(vector<mpol_t> &E)
{
	HASH_CELL ** table;
	long ind=0;
	bool first=true;

	lmono.clear();
	table=get_monomial(E);
	R.clear();

	if (verbose) {
		cout << "Adding " << lmono.size() << " new variables." << endl;
		for (int i=0; i<lmono.size(); i++) {
			cout << *lmono.at(i) << " y" << i << endl; 
		}
	}


	M.resize(E.size(),lmono.size()+vars.size());
	for (int i=0; i<M.nbrow(); i++)
		for (int j=0; j<M.nbcol(); j++)
			M(i,j)=0;

	R.resize(E.size(),0);

	for (int i=0; i<E.size(); i++) {
		first=true;
		for(mpol_t::iterator it=E.at(i).begin(); 
			it !=E.at(i).end(); it++) {
			if (deg(it) > 0) {	// not constant
				ind=uni_check(it);
				if (ind==-1) {
					M(i,mono_member(it))+=it->coeff();
				} else	{
					M(i,lmono.size()+ind-1)+=it->coeff();
				}	
			} else { // constant
				R.at(i)+=it->coeff();
			}
		}	
	}

	ht_free(table, ht_size);
}

/*
 * Returns -1 if monomial is not univariate or returns index of var
 */
long
uni_check(mpol_t::iterator mono)
{
	long first=0;

	// Find first variable
	for (long i=0; i<mono->size(); i++) {
		if ((*mono)[i]>0) {
			first=++i;
			break;
		}
	}
	for (long i=first; i<mono->size(); i++) 
		if ((*mono)[i]>0) return -1; 

	return first;
}

/*
 * Gauss elimination for dense matrix
 * works only for GF(2)
*/
unsigned long
g2elim(matrix_t &A, vector<coeff_t> &B)
{
	int i=0, j=0;
	unsigned long rank = 0; 
	int m=A.nbrow();
	int n=A.nbcol();
	coeff_t tmp,max,val;
	int ind=0;

	while ((i < m) && (j < n)) {
		max=A(i,j);
		ind=i;
		for (int k=i+1; k<m; k++) {
			val=A(k,j);
			if (val > max) {
			max=val;
			ind=k;
			}
		}
		if (max != 0) {
			rank++;
			A.swaprow(i, ind);
			tmp=B[i];
			B[i]=B[ind];
			B[ind]=tmp;
			for (int k=0; k<m; k++) {
				if ((k!=i) && (A(k,j)!=0)) {
					A.addrow(k, i, 1);
					B[k]+=B[i];
				}
			}
			i++;
		}
		j++;
	}
	return rank;
}

/*
 * Load all equations from input file.
 * format left side = right side
 */
void
load_eq(char * input, vector<mpol_t> &E)
{
	fstream in(input, ios::in);
	if (in.is_open()){
		string line, ls, rs;
		int ind;
		while (!in.eof()) {
			getline(in, line,'\n');
			if ( (string::npos != (ind=line.find_first_of("=", 0))) and (line.at(0) != '%') ) {
				ls=line.substr(0, ind);
				rs=line.substr(ind+1, string::npos);
				mpol_t left(ls.c_str());
				mpol_t right(rs.c_str());
				nin+=left.size()-1;
				if (right != 0)
					nin++;
				if ( (left.size()-1 > maxin) && (right != 0) )
					maxin=left.size();
				if ( (left.size()-1 > maxin) && (right == 0) )
					maxin=left.size();
				mpol_t result = left -right;
				// get list of variables
				init_vars(result);
				E.push_back(result);
			}
		}
		// set vector of vars
		for (set<string>::iterator i=svars.begin(); i!=svars.end(); i++) { 
			vars.push_back(*i);
			occured.push_back(false);
		}
		// Since this point, I will use only vector of all vars
		svars.clear(); 
	}
  	else {
		cerr << "ERROR: Unable to open input file " << input << endl;
  		exit(1);
	}
	in.close();
}

void
print_res(char * output)
{
	mpol_t poly=0;
	unsigned long counter=0;
	long nvar=vars.size(), cols=M.nbcol();
	fstream out(output, ios::out);

	if ( out.good() ) {	
		for (int i=0; i<M.nbrow(); i++) {
			poly = 0;
			for (int j=0; j<cols; j++) {
				if (M(i,j) != 0) {
					if (j < cols-nvar) {
						poly += *lmono.at(j);
						for (int k=0; k<=lmono.at(j)->nvars(); k++) {
							if (lmono.at(j)->rep()[k] != 0) {
								occured.at(k)=true;	
							}
						}
					} else {
						mpol_t foo(vars.at(j-cols+nvar).c_str()); 
						poly += foo;
						occured.at(j-cols+nvar)=true;
					}
				}
			}
			nout+=poly.size()-1;
			if (poly.size()-1 > maxout)
				maxout = poly.size();
			out << poly << "=" << R[i] << endl;	
			counter++;
			if ((necc_eqs == 0) && really_true(occured)) {
				necc_eqs=counter;
			}
		}
	} else {
		cerr << "ERROR: Unable to open output file " << output << endl;
		exit(1);
	}
}

/*
 * Insert all vars of poly into set.
 */
void
init_vars(mpol_t & a)
{
	for(mpol_t::iterator it =a.begin(); it !=a.end(); it++) {
		for (int i=0; i<=it->nvars(); i++) {
			if ( it->rep()[i] > 0) {
				svars.insert(it->var[i]);	
			}
		}
	}
}

void
arguments(int argc, char* argv[])
{
	int next;
	const char* const short_options = "hvVH:";
	do {
	next = getopt(argc, argv, short_options);
	switch (next) {
		case '?' : usage(); break;
		case 'V' : cout <<"elim (" << MYVERSION << ")" << endl;
			   exit(1);
			   break;
		case 'v' : verbose=true; 
			   break;
		case 'H' : ht_size=atol(optarg); 
			   break;
		case 'h' : usage(); 
			   exit(1);
			   break;
	}
	} while (next != -1 );
	if (optind != argc) input=argv[optind];
	if (++optind != argc) output=argv[optind];
	/*Value chceck*/
	if ((input == NULL) || (output == NULL)) {
		cerr << "ERROR: Required parameter is missing." << endl;
		usage();
		exit(1);
	} 
}

void 
usage ()
{
	cout <<"elim (" << MYVERSION << ")" << endl;
	cout <<"Usage: elim [options] input output"<< endl;
	cout <<"	-h : this help message" << endl
		<<"	-V : print version info" << endl
		<<"	-v : verbose mode " << endl
		<<"	-H : size of hash table [default " << ht_size << "]" << endl;
}

