/*
Program: xl
File: xl.cpp
Author: Martin Voros
*/

/*
NOTES:
!!! all variables in equations have to be in format xnum. 
*/

#include <iostream>
#include <fstream>
#include <vector>
#include <iterator>
#include <getopt.h>


/*SYNAPS headers
#include "synaps/mpol.h"
#include "synaps/arithm/Zp.h"
#include "synaps/linalg.h"
*/

#include "xl.h"
#include "hash.h"
#include "comb.h"

/*Global vars*/
char * input=NULL;
bool verbose;
short mode = 0;
int D = 4;
int rlimit = 0;
int dmax = 0; // maximal degree
unsigned long ht_size=HT_DEF_SIZE;
bool stat=false;

// Set & vector of all variables
set<string> svars;
vector<string> vars;
// Vector of the solutions, -1 means unknown
vector<int> sol;
// Equations
vector<mpol_t> eqs;
vector<mpol_t> eqs_w;
matrix_t M;
vector<coeff_t> R;
// List of all monomials
vector<mpol_t::iterator> lmono;
// All combinations with repetition
ListComb allcomb;

using namespace std;

int 
main (int argc, char**argv) 
{
	/* Stats */
	int uvars; // # of unknown variables
	int rounds = 0;
	vector<long> Rfree;
	vector<long> Treal;

	long var_index=0;
	bool status=true;
	unsigned long rank;
	mpol_t poly=0;

	/*Parsing input parameters*/
	arguments(argc, argv);
	load_eq(input, eqs);
	uvars=vars.size();
	if (rlimit == 0)
		rlimit = uvars;
	if (mode >0) {
		if (mode == 1) 
			for (int i=2; i<=(D-dmax); i+=2) gen_comb(vars.size(), i, allcomb);
		if (mode == 2) 
			for (int i=1; i<=(D-dmax); i+=2) gen_comb(vars.size(), i, allcomb);
		if (mode == 3) 
			gen_comb(vars.size(), D-dmax, allcomb);
	} else {
		for (int i=1; i<=(D-dmax); i++) 
			gen_comb(vars.size(), i, allcomb);
	}

	for (int i=0; i<eqs.size(); i++) {
			eqs_w.push_back(eqs.at(i));
	}

/*
 *	XL Algorithm
 */
	while (status && (uvars>0) && (rlimit>0)) {
		rlimit--;
		rounds++;
		// Step 1
		generate_product(eqs_w);
		reduce_equations(eqs_w, D);

		if (verbose) {
			cout << "[ --- System of equations (" << eqs_w.size() << ") --- ]" << endl;
			print_eqs(eqs_w);
		}

		// Step 2
		linearize(eqs_w);
		// Solve system
		rank=g2elim(M,R);
		
		if (stat) {
			Treal.push_back(M.nbcol());
			Rfree.push_back(rank);
		}
		
		if (verbose) {
			cout << "After elimination"<<endl;
			cout << "Matrix M:" << endl;
			cout << M;
			cout << "Vector R:" << endl;
			for (int i=0; i<R.size(); i++)
				cout << R.at(i) << " ";
 			cout << endl;
		}

		// Step 3
		// Find and solve univariate polynomial
		status=false;
		var_index=0;
		coeff_t solution=0;
		for (int i=0; i<M.nbrow(); i++) {
			if ( (var_index=uni_check2(i)) != -1) {
				status=true;
				solution=solve_brutegf2(i);
				remove_comb(allcomb,var_index);
				sol.at(var_index)=solution.rep();
				uvars--;
			}
		}

						
		if (verbose)
			print_sol();

		// Substitute solution
		eqs_w.clear();
		for (int i=0; i<eqs.size(); i++) {
			eval_poly(eqs.at(i),poly, sol);
			if (poly != 0)
				eqs_w.push_back(poly);
		}
				
	
	} // End of XL

	if (!verbose)
		print_sol();

	if (stat) {
		cout << "[ --- Stats --- ]" << endl;
		cout << "Rounds: " << rounds << endl;
		cout << "Unknown vars: " << uvars << endl;
		cout << "# of vars: " << vars.size() << endl;
		cout << "# of polynomials: " << eqs.size() << endl;
		cout << "Max. degree of eqs: " << dmax << endl;
		cout << "parameter D=" << D << endl;
		cout << "T_real: ";
		for (int i=0; i<Treal.size(); i++)
			cout << "T(" << i << ")=" << Treal.at(i) << " ";
		cout << endl;
		cout << "R_free: ";
		for (int i=0; i<Rfree.size(); i++)
			cout << "R(" << i << ")=" << Rfree.at(i) << " ";
		cout << endl;
	}

	return 0;
}

/*
 * Evaluate polynomial using values from sol
 */
void
eval_poly(mpol_t &m, mpol_t &r, vector<int> &sol)
{
	bool zero;
	r=0;

	for(mpol_t::iterator it=m.begin(); it !=m.end(); it++) {
		zero=false;
		for (int j=0; j<it->nvars(); j++) {
			if ( (it->rep()[j] > 0) &&  (sol.at(j) == 0) ) {
				it->coeff()=0;
				zero=true;
				break; // for 
			}  
			if ( (it->rep()[j] > 0) &&  (sol.at(j) == 1) ) {
				it->rep()[j]=0;
			} 
		}
		if (!zero)
			r += *it;
	}
}


/*
 * Brute force solver for univariate polynomial over GF(2)
 */
coeff_t
solve_brutegf2(long row)
{
	coeff_t val=eval_mrow(row,0);
	if (val == R[row])
		return 0;
	else
		return 1; 
}

/*
 * Eval multivariate poly - substitute all variables by val
 */
coeff_t
eval_mrow(long row, coeff_t val)
{
	long n=M.nbcol();
	long border=n-vars.size();
	mpol_t::iterator mono;
	coeff_t result=0, tmp=0;

	for (int j=0; j<n; j++)
		if  (M(row,j)!=0) {
			if (j<border) {
				mono=lmono.at(j);
				tmp=mono->coeff();
				for (long i=0; i<mono->size(); i++) 
					tmp = tmp * val ^ ((*mono)[i]); 
			} else {
				tmp=val;	
			}
			result+=tmp;	
		}

	return result;
}

/*
 * Returns true when polynom is univariate
 */
long
uni_check2(long row)
{
	long n=M.nbcol();
	long border=n-vars.size();
	long p=-1,index=-1;
	long first;
	
	/* Find first variable */
	// Find first non-zero value in row
	for (int j=0; j<n; j++)
		if  (M(row,j)!=0) {
			p=j;
			break;
		}
	if (p==-1)
		// zero row
		return -1;
	else {
		if (p<border) {
			index=uni_check(lmono.at(p));
			if (index!=-1) {
				first=index;
			} else {
				return -1;	
			}
		} else {
			first=p-border; 
		}
	}
	/* check other monomials */
	mpol_t::iterator mono;
	for (int j=p+1; j<n; j++) {
		if  (M(row,j)!=0) {
			if (j<border) {
				// y_n
				mono=lmono.at(j);
				for (long i=0; i<mono->size(); i++) 
					if ( ((*mono)[i]>0) && (i!=first) ) 
						return -1; 
			} else {
				// x_n
				if ( (j-border) != first ) 
					return -1;
			}
		}
	}
	
	// cout << "Variable: " << vars.at(first) << endl;
	return first; 
}

/*
 * Returns index of monomial in list of the monos else -1
 */
long
mono_member(mpol_t::iterator mono)
{
	for (int i=0; i<lmono.size(); i++) {
		if (*mono == *lmono.at(i)) {
			return i;
		}
	}
	return -1;
}

/*
 * Returns -1 if monomial is not univariate or returns index of var
 */
long
uni_check(mpol_t::iterator mono)
{
	long first=0;

	// Find first variable
	for (long i=0; i<mono->size(); i++) {
		if ((*mono)[i]>0) {
			first=++i;
			break;
		}
	}
	for (long i=first; i<mono->size(); i++) 
		if ((*mono)[i]>0) return -1; 

	return first;
}


/*
 * Returns degree of monomial
 */
int
deg(mpol_t::iterator mono)
{
	long deg=0;

	for (long i=0; i<mono->size(); i++) 
		deg+=(*mono)[i];

	return deg;
}


/*
 * Get list of all monomials
 */
HASH_CELL **
get_monomial(vector<mpol_t> &eqs)
{
	HASH_CELL ** table;
	unsigned long key;
	unsigned long id=1;
	
	table=ht_alloc(ht_size);
	for (int i=0; i<eqs.size(); i++) {
		for(mpol_t::iterator it=eqs.at(i).begin(); 
			it !=eqs.at(i).end(); it++) {
			if (deg(it)>1) {
				key=hash_mono(it,ht_size);
				if ( 0 == hash_lookup(it, table[key]) ) {
					table[key]=insert_mono(it, table[key], lmono, id);
					id++;
				}
			}
		}	
	}

	return table;
}

/*
 * Linearize system of equations
 */
void
linearize(vector<mpol_t> &E)
{
	HASH_CELL ** table;
	long ind=0;

	lmono.clear();
	table=get_monomial(E);
	R.clear();

	if (verbose) {
		cout << "Adding " << lmono.size() << " new variables." << endl;
		for (int i=0; i<lmono.size(); i++) {
			cout << *lmono.at(i) << " y" << i << endl; 
		}
	}

	M.resize(E.size(),lmono.size()+vars.size());
	for (int i=0; i<M.nbrow(); i++)
		for (int j=0; j<M.nbcol(); j++)
			M(i,j)=0;
	R.resize(E.size(),0);

	for (int i=0; i<E.size(); i++) {
		for(mpol_t::iterator it=E.at(i).begin(); 
			it !=E.at(i).end(); it++) {
			if (deg(it) > 0) {	// not constant
				ind=uni_check(it);
				if (ind==-1) {
					M(i,mono_member(it))+=it->coeff();
				} else	{
					M(i,lmono.size()+ind-1)+=it->coeff();
				}	
			} else { // constant
				R.at(i)+=it->coeff();
			}
		}	
	}

	ht_free(table, ht_size);

	if (verbose) {
		cout << "Matrix M:" << endl;
		cout << M;
		cout << "Vector R:" << endl;
		for (int i=0; i<R.size(); i++)
			cout << R.at(i) << " ";
 		cout << endl;
	}
}


/*
 * Gauss elimination for dense matrix
 * works only for GF(2)
*/
unsigned long
g2elim(matrix_t &A, vector<coeff_t> &B)
{
	int i=0, j=0;
	unsigned long rank = 0; 
	int m=A.nbrow();
	int n=A.nbcol();
	coeff_t tmp,max,val;
	int ind=0;

	while ((i < m) && (j < n)) {
		max=A(i,j);
		ind=i;
		for (int k=i+1; k<m; k++) {
			val=A(k,j);
			if (val > max) {
			max=val;
			ind=k;
			}
		}
		if (max != 0) {
			rank++;
			A.swaprow(i, ind);
			tmp=B[i];
			B[i]=B[ind];
			B[ind]=tmp;
			for (int k=0; k<m; k++) {
				if ((k!=i) && (A(k,j)!=0)) {
					A.addrow(k, i, 1);
					B[k]+=B[i];
				}
			}
			i++;
		}
		j++;
	}
	return rank;
}

/*
 * Add new equations
 */
void
generate_product(vector<mpol_t> &E)
{
	vector<mpol_t> new_eqs;

	for (int j=0; j<E.size(); j++) {
		for (int i=0; i<allcomb.size(); i++) {
			mpol_t res("1");
			for (int k=0; k<allcomb.at(i).size(); k++) { 
				mpol_t foo(vars.at(allcomb.at(i).at(k)).data()); 
				res = foo * res;
			}
			new_eqs.push_back(E.at(j) * res);
		}
	}
	for (int i=0; i<new_eqs.size(); i++)
		E.push_back(new_eqs.at(i));
}

/*
 * Discards equations with degree > max
 */
void
reduce_equations(vector <mpol_t> &E, int max)
{
	int deg = 0;

	for (vector<mpol_t>::iterator i=E.begin(); i!=E.end(); i++) {
		for(mpol_t::iterator it=i->begin(); it !=i->end(); it++) {
			deg = 0;
			for (int j=0; j<=it->nvars(); j++) {
				if ( ((it->rep()[j] % FIELD) == 0) && (it->rep()[j] != 0) ) {
					it->rep()[j]=1;
				} else 
					it->rep()[j]%=FIELD;
				deg+=it->rep()[j];  
			}
			/* Discard equatiosn with degree > D */
			if ( deg > max) {
				E.erase(i);
				if (i != E.begin())
					i--;
				break;
			} 
		}
	}
}

/*
 * Load all equations from input file.
 * format left side = right side
 */
void
load_eq(char * input, vector<mpol_t> &E)
{
	fstream in(input, ios::in);
	if (in.is_open()){
		string line, ls, rs;
		int ind;
		while (!in.eof()) {
			getline(in, line,'\n');
			if ( (string::npos != (ind=line.find_first_of("=", 0))) and (line.at(0) != '%') ) {
				ls=line.substr(0, ind);
				rs=line.substr(ind+1, string::npos);
				mpol_t left(ls.c_str());
				mpol_t right(rs.c_str());
				mpol_t result = left -right;
				if ( degree(result) > dmax)
					dmax=degree(result);
				// get list of variables
				init_vars(result);
				E.push_back(result);
			}
		}
		// set vector of vars
		for (set<string>::iterator i=svars.begin(); i!=svars.end(); i++) { 
			vars.push_back(*i);
			sol.push_back(-1);
		}
		// Since this point, I will use only vector of all vars
		svars.clear(); 
	}
  	else {
		cerr << "ERROR: Unable to open input file " << input << endl;
  		exit(1);
	}
	in.close();

	if (verbose) {
		cout << "[ --- Initial equations (" << E.size() << ") --- ]" << endl;
		print_eqs(E);
		cout << "List of all variables (" << vars.size() << "):" << endl;
		for (int i=0; i<vars.size(); i++) {
			cout << vars.at(i) << " ";
		}
		cout << endl;
	}
}

/*
 * Insert all vars of poly into set.
 */
void
init_vars(mpol_t & a)
{
	for(mpol_t::iterator it =a.begin(); it !=a.end(); it++) {
		for (int i=0; i<=it->nvars(); i++) {
			if ( it->rep()[i] > 0) {
				svars.insert(it->var[i]);	
			}
		}
	}
}

void 
print_eqs(vector<mpol_t> &eqs) 
{
	for (int i=0; i<eqs.size(); i++) {
		cout << eqs.at(i) << endl;
	}
}

/*
 * Print solution
 */
void 
print_sol() 
{
	cout << "[ --- Solution --- ]" << endl;
	for (int i=0; i<sol.size(); i++) {
		cout << vars.at(i) << "=";
		if (sol.at(i) == -1) 
			cout << "?" << endl;
		else
			cout << sol.at(i) << endl;
	}
}

void
arguments(int argc, char* argv[])
{
	int next;
	const char* const short_options = "D:hH:m:r:svV";
	do {
	next = getopt(argc, argv, short_options);
	switch (next) {
		case '?' : usage(); break;
		case 'V' : cout <<"xl (" << MYVERSION << ")" << endl;
			   exit(1);
			   break;
		case 'm' : if (strncmp(optarg,"a",1) == 0) {
				mode = 0;
			   } 
			   if (strncmp(optarg,"e",1) == 0) {
				mode = 1;
			   } 
			   if (strncmp(optarg,"o",1) == 0) {
				mode = 2;
			   }
			   if (strncmp(optarg,"l",1) == 0) {
				mode = 3;
			   } 
			   break;	
		case 'r' : rlimit=atol(optarg); break;
		case 's' : stat=true; break;
		case 'v' : verbose=true; break;
		case 'D' : D=atol(optarg); 
			   if (D < 3) D=4;
			   break;
		case 'H' : ht_size=atol(optarg); 
			   break;
		case 'h' : usage(); 
			   exit(1);
			   break;
	}
	} while (next != -1 );
	if (optind != argc) input=argv[optind];
	/*Value chceck*/
	if (input == NULL) {
		cerr << "ERROR: Required parameter is missing." << endl;
		usage();
		exit(1);
	} 
}

void 
usage ()
{
	cout <<"xl (" << MYVERSION << ")" << endl;
	cout <<"Usage: xl [options] equations"<< endl
		<<"	-h : this help message" << endl
		<<"	-V : print version info" << endl
		<<"	-v : verbose mode" << endl
		<<"	-D : D (>3) parameter [default 4]" << endl
		<<"	-m : generate x^k*l for " << endl
		<<"		a - k=1..D-dmax [default]" << endl
		<<"		e - k=1..D-dmax & k is even" << endl
		<<"		o - k=1..D-dmax & k is odd" << endl
		<<"		l - k=D-dmax" << endl
		<<"	-H : size of hash table [default " << ht_size << "]" << endl
		<<"	-r : maximum # of iterations [default # of vars]" << endl
		<<"	-s : print stat"
		<<endl;
}

